

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
    *                                                                 *
    *   DEVIL Screensaver System 1.05                                 *
    *   Copyright  2005 Sebastian Pipping <webmaster@hartwork.org>   *
    *                                                                 *
    *   --> http://www.hartwork.org                                   *
    *                                                                 *
    *                                                                 *
    *   This source code is released under LGPL.                      *
    *   See LGPL.txt for details.                        2005-09-18   *
    *                                                                 *
    \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#ifndef DEVIL_H
#define DEVIL_H


#include "DevilGlobal.h"


enum DEVIL_MODE
{
	DEVIL_MODE_UNKNOWN,			// Before detection

	DEVIL_MODE_CONFIG,
	DEVIL_MODE_PASSWORD,
	DEVIL_MODE_FULLSCREEN,
	DEVIL_MODE_PREVIEW,
	DEVIL_MODE_WINDOWED,		// For debugging purpose...
};



typedef bool ( * DEVIL_FUNC_LIVE )( void );
typedef void ( * DEVIL_FUNC_DIE )( void );
typedef void ( * DEVIL_FUNC_RENDER )( const double );
typedef void ( * DEVIL_FUNC_CONFIG )( const HWND );

typedef bool ( * DEVIL_FUNC_INIT_3D )( void );
typedef void ( * DEVIL_FUNC_CLEANUP_3D )( void );
typedef bool ( * DEVIL_FUNC_SWITCH_RES )( int *, int *, UINT *, UINT * );
typedef void ( * DEVIL_FUNC_RESTORE_RES )( void );



struct DEVIL_INTERFACE
{
	// -------------------- Function pointers ------------------------------------------
	DEVIL_FUNC_LIVE			Live;					// Non-3D init stuff goes here:
													// Load settings, check windows version, ...
													// Return [false] to quit
													// Optional = can be NULL

	DEVIL_FUNC_DIE			Die;					// Non-3D cleanup stuff goes here:
													// Save seetings, delete tempfiles, ...
													// Optional = can be NULL
	
	WNDPROC					Master;					// Master window procedure
													// Optional = can be NULL

	DEVIL_FUNC_RENDER		Render;					// Called from inside the message loop
													// whenever no message is waiting
													// Optional = can be NULL

	DEVIL_FUNC_CONFIG		Config;					// Configuraion dialog goes here
													// If this is NULL the default config is used
													// See [DEVIL_DefaultConfig]
													// Optional = can be NULL

	DEVIL_FUNC_INIT_3D		Init3D;					// Called between window creation and message loop
													// OpenGL/D3D init stuff goes here.
													// [hwnd] member is now valid
													// Return [false] to quit
													// Optional = can be NULL

	DEVIL_FUNC_CLEANUP_3D	Cleanup3D;				// Called between message loop and window destruction
													// Optional = can be NULL
	
	DEVIL_FUNC_SWITCH_RES	SwitchRes;				// Called right before window creation
													// Switch screen resolution here and
													// adjust window dimensions if needed
													// Multi-mon-handling goes here, too
													// Return [false] to quit
													// Optional = can be NULL

	DEVIL_FUNC_RESTORE_RES	RestoreRes;				// Restore original screen resolution here
													// Called right after window destruction
													// Optional = can be NULL


	// -------------------- Frame limiter ----------------------------------------------
	bool					bEnableFrameLimiter;	// Limit FPS to [uMaxFPS] ON/OFF
													// Can be changed while running

	UINT					uMaxFPS;				// Maximum FPS if limiter is active
													// Can be changed while running


	// -------------------- Behavior of [Slave] ----------------------------------------
	bool					bMouseExit;				// Mouse action quits saver ON/OFF.
													// Can be changed while running

	bool					bAnyKeyExit;			// Any keystroke quits saver ON/OFF
													// If this is off only escape quits
													// Can be changed while running


	// ==================== SET BY HOST ================================================
	HWND					hwnd;					// Saver window hanlde
													// This member is filled right before
													// [Init3D] is called.

	DEVIL_MODE				Mode;					// Mode in which the saver will run
													// See definition of DEVIL_MODE

	WNDPROC					Slave;					// Slave window procedure
													// Master has to pass to this

	double					fFPS;					// Current FPS
													// Refreshed before each frame
};



/*

Functions are called in this order:


            Live
             |
      /------+------\
      |             |
     Config        SwitchRes
      |             |
      |            Init3D
      |             |
      |            <LOOP> Master, Render
      |             |
      |            Cleanup3D
      |             |
      |            RestoreRes
      |             |
      \------+------/
             |
		    Die

*/


#endif // DEVIL_H
