

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
    *                                                                 *
    *   DEVIL Screensaver System 1.05                                 *
    *   Copyright  2005 Sebastian Pipping <webmaster@hartwork.org>   *
    *                                                                 *
    *   --> http://www.hartwork.org                                   *
    *                                                                 *
    *                                                                 *
    *   This source code is released under LGPL.                      *
    *   See LGPL.txt for details.                        2005-09-18   *
    *                                                                 *
    \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "DevilGlobal.h"
#include "Devil.h"
#include "DevilTimer.h"
#include <stdlib.h>
#include <process.h>


#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL 0x020A
#endif


#define DEVIL_APPTITLE		TEXT( "DEVIL Screensaver System" )
#define DEVIL_WINDOWCLASS	TEXT( "DEVILCLASS" )


#define MIN_MOUSEMOVE 10



typedef DWORD ( FAR PASCAL * PWCHGPROC )( LPCTSTR, HWND, DWORD, LPVOID );
typedef BOOL ( FAR PASCAL * VERIFYPWDPROC )( HWND );



// General
extern DEVIL_INTERFACE g_FACE;
HWND g_hParent = NULL;
HINSTANCE g_hInstance = NULL;
bool g_bActive = true;


// Mouse
POINT g_CursorPosCur;
POINT g_CursorPosFirst;


// Password
HINSTANCE g_hPWD = NULL;
VERIFYPWDPROC VerifyPassword = NULL;

bool g_bFullscreenQuitPermission  = true;
bool g_bFullscreenVerifying       = false;



LRESULT CALLBACK DEVIL_Slave( HWND hwnd, UINT message, WPARAM wp, LPARAM lp );



// *****************************************************************************
bool DEVIL_GetMode()
{
	LPTSTR szCmdLine = GetCommandLine();

	// Skip filename
    if( *szCmdLine == TEXT( '\"' ) )
	{
		szCmdLine++;
		while( *szCmdLine != TEXT( '\"' ) )
		{
			szCmdLine++;
		}
        if ( *szCmdLine == TEXT( '\"' ) )
		{
            szCmdLine++;
		}
	}
	else
	{
		while( *szCmdLine != TEXT( ' ' ) )
		{
			szCmdLine++;
		}
	}

	// Skip spaces
	while( *szCmdLine == TEXT( ' ' ) )
	{
		szCmdLine++;
	}

	while( true )
	{
		switch( *szCmdLine )
		{
		case TEXT( 'S' ):
		case TEXT( 's' ):
			g_FACE.Mode  = DEVIL_MODE_FULLSCREEN;
			g_hParent    = NULL;

			return true;

		case TEXT( 'W' ):
		case TEXT( 'w' ):
			g_FACE.Mode  = DEVIL_MODE_WINDOWED;
			g_hParent    = NULL;

			return true;

		case TEXT( 'P' ):
		case TEXT( 'p' ):
			szCmdLine++;

			// Skip spaces
			while( *szCmdLine == TEXT( ' ' ) )
			{
				szCmdLine++;
			}

			g_FACE.Mode	= DEVIL_MODE_PREVIEW;
			g_hParent	= ( HWND )_ttoi( szCmdLine );

			if( IsWindow( g_hParent ) )
			{
				return true;
			}
			else
			{
				g_hParent = GetForegroundWindow();
				return false;
			}

		case TEXT( 'A' ):
		case TEXT( 'a' ):
			szCmdLine++;

			// Skip spaces
			while( *szCmdLine == TEXT( ' ' ) )
			{
				szCmdLine++;
			}

			g_FACE.Mode = DEVIL_MODE_PASSWORD;

			if( *szCmdLine != TEXT( ' ' ) )
			{
				g_hParent = ( HWND )_ttoi( szCmdLine );
				if( !IsWindow( g_hParent ) )
				{
					g_hParent = GetForegroundWindow();
				}
			}
			else
			{
				g_hParent = GetForegroundWindow();
			}

			return true;

		case TEXT( 'C' ):
		case TEXT( 'c' ):
			szCmdLine++;

			// Skip spaces
			while( *szCmdLine == TEXT( ' ' ) )
			{
				szCmdLine++;
			}

			g_FACE.Mode = DEVIL_MODE_CONFIG;

			if( *szCmdLine != TEXT( ' ' ) )
			{
				g_hParent = ( HWND )_ttoi( szCmdLine );
				if( !IsWindow( g_hParent ) )
				{
					g_hParent = GetForegroundWindow();
				}
			}
			else
			{
				g_hParent = GetForegroundWindow();
			}

			return true;

		case TEXT( ' ' ):
		case TEXT( '-' ):
		case TEXT( '/' ):
			szCmdLine++;
			break;

		default:
			g_FACE.Mode	= DEVIL_MODE_CONFIG;
			g_hParent	= GetForegroundWindow();

			return true;

		}
	}
}



// *****************************************************************************
bool DEVIL_LoadPWD()
{
	g_hPWD = LoadLibrary( TEXT( "PASSWORD.CPL" ) );
	if( g_hPWD != NULL )
	{
		VerifyPassword = ( VERIFYPWDPROC )GetProcAddress( g_hPWD, "VerifyScreenSavePwd" );	// ANSI !!!
	}

	if( VerifyPassword == NULL )
	{
		FreeLibrary( g_hPWD );
		return false;
	}
	else
	{
		return true;
	}
}



// *****************************************************************************
void DEVIL_UnloadPWD()
{
	if( g_hPWD != NULL )
	{
		FreeLibrary( g_hPWD );
	}
}



// *****************************************************************************
bool DEVIL_CheckInterface()
{
	g_FACE.Slave = DEVIL_Slave;

	return true;
}



// *****************************************************************************
void DEVIL_QueryPasswordThread( PVOID pvoid )
{
	ShowCursor( TRUE );

	// Show dialog
	g_bFullscreenQuitPermission = !!VerifyPassword( g_FACE.hwnd );

	g_bFullscreenVerifying = false;
	GetCursorPos( &g_CursorPosFirst );
	ShowCursor( FALSE );

	if( g_bFullscreenQuitPermission )
	{
		PostMessage( g_FACE.hwnd, WM_CLOSE, 0, 0 );
	}
}



// *****************************************************************************
void DEVIL_ChangePassword()
{
	HINSTANCE mpr = LoadLibrary( TEXT( "MPR.DLL" ) );
	if( mpr != NULL )
	{
#ifndef UNICODE
		PWCHGPROC pwd = ( PWCHGPROC )GetProcAddress( mpr, "PwdChangePasswordA" );	// ANSI !!!
#else
		PWCHGPROC pwd = ( PWCHGPROC )GetProcAddress( mpr, "PwdChangePasswordW" );	// ANSI !!!
#endif

		if( pwd != NULL )
		{
			pwd( TEXT( "SCRSAVE" ), g_hParent, 0, NULL );
		}

		FreeLibrary( mpr );
	}
}



// *****************************************************************************
void DEVIL_DefaultConfig()
{
	MessageBoxEx(
		g_hParent,
		TEXT( "This screensaver was created with DEVIL Screensaver System.  \n"
			"Download it now and write your own screensaver today.\n"
			"\n"
			"-->  http://www.hartwork.org\n" ),
		DEVIL_APPTITLE,
		MB_ICONINFORMATION,
		MAKELANGID( LANG_ENGLISH, SUBLANG_ENGLISH_US )
	);
}



// *****************************************************************************
LRESULT CALLBACK DEVIL_Slave( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
{
	static bool bDelayFixed = false;
	
	switch( message )
	{
	case WM_SHOWWINDOW:
		// Delay workaround
		if( wp && !bDelayFixed && ( g_FACE.Mode == DEVIL_MODE_PREVIEW ) )
		{
			HWND DisplayPropteries = GetParent( g_FACE.hwnd );
			DisplayPropteries = GetParent( DisplayPropteries );
			DisplayPropteries = GetParent( DisplayPropteries );
			DisplayPropteries = GetParent( DisplayPropteries );
			SetForegroundWindow( DisplayPropteries );
			bDelayFixed = true;
		}
		break;

	case WM_SIZE:
		// Pause rendering when minimized
		if( g_FACE.Mode == DEVIL_MODE_WINDOWED )
		{
			switch( wp )
			{
				case SIZE_MAXHIDE:
				case SIZE_MINIMIZED:
					g_bActive = false;
					break;
					
				default:
					g_bActive = true;
			}
		}
		break;

	case WM_CREATE:
		// Hide cursor in fullscreen
		if( g_FACE.Mode == DEVIL_MODE_FULLSCREEN )
		{
			GetCursorPos( &g_CursorPosFirst );
			ShowCursor( FALSE );

			return FALSE;
		}
		break;

	case WM_KEYDOWN:
		if( ( g_FACE.Mode == DEVIL_MODE_FULLSCREEN ) && !g_bFullscreenVerifying )
		{
			if( g_FACE.bAnyKeyExit || ( ( wp == VK_ESCAPE ) && !( lp & 0x40000000 ) ) )
			{
				PostMessage( g_FACE.hwnd, WM_CLOSE, 0, 0 );
			}
			return 0;
		}
		break;

	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_MOUSEWHEEL:
		if( ( g_FACE.Mode == DEVIL_MODE_FULLSCREEN ) && !g_bFullscreenVerifying && g_FACE.bMouseExit )
		{
			PostMessage( g_FACE.hwnd, WM_CLOSE, 0, 0 );
			return 0;
		}
		break;

	case WM_MOUSEMOVE:
		if( ( g_FACE.Mode == DEVIL_MODE_FULLSCREEN ) && !g_bFullscreenVerifying && g_FACE.bMouseExit )
		{
			GetCursorPos( &g_CursorPosCur );

			// Accuracy is not needed here so we don't check the real distance
			// which we would need [sqrt] for but just the distance on x- and y-axis
			// which is much faster...
			if(	( abs( g_CursorPosCur.x - g_CursorPosFirst.x ) >= MIN_MOUSEMOVE ) ||
				( abs( g_CursorPosCur.y - g_CursorPosFirst.y ) >= MIN_MOUSEMOVE ) )
			{
				PostMessage( g_FACE.hwnd, WM_CLOSE, 0, 0 );
			}

			return 0;
		}
		break;

	case WM_SYSCOMMAND:
		if( g_FACE.Mode == DEVIL_MODE_FULLSCREEN )
		{
			switch( wp )
			{
			case SC_CLOSE:			// Deny [ALT]+[F4]
			case SC_SCREENSAVE:		// Deny another screensaver start
				return 0;

			}
		}
		break;

	case WM_CLOSE:
		if( g_FACE.Mode == DEVIL_MODE_FULLSCREEN )
		{
			if( !g_bFullscreenQuitPermission && !g_bFullscreenVerifying && ( VerifyPassword != NULL ) )
			{
				// Start additional thread
				g_bFullscreenVerifying = true;
				_beginthread( DEVIL_QueryPasswordThread, 0, NULL );
				return 0;		// Deny closing
			}
		}
		break;

	case WM_DESTROY:
		PostQuitMessage( 0 );
		return 0;

	}
	return DefWindowProc( hwnd, message, wp, lp );
}



// *****************************************************************************
int DEVIL_RunSaver()
{
	// Disable some keys...
	if( g_FACE.Mode == DEVIL_MODE_FULLSCREEN )
	{
#ifndef _DEBUG
		SystemParametersInfo( SPI_SCREENSAVERRUNNING, TRUE, NULL, 0 );
#endif

		OSVERSIONINFO osvi;
		osvi.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
		GetVersionEx( &osvi );
		if( osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS )
		{
			DEVIL_LoadPWD();
		}
	}


	// Register class
	WNDCLASS wc;
	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc		= ( ( g_FACE.Master != NULL ) ? g_FACE.Master : DEVIL_Slave );
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= g_hInstance;
	wc.hIcon			= LoadIcon(NULL, IDI_WINLOGO );
	wc.hCursor			= LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground	= ( HBRUSH )GetStockObject( BLACK_BRUSH );	// HOLLOW_BRUSH
	wc.lpszMenuName		= NULL;
	wc.lpszClassName	= DEVIL_WINDOWCLASS;

	if( !RegisterClass( &wc ) )
	{
		return 0;
	}


	// Mode-dependent window creation settings
	DWORD dwExStyle;
	DWORD dwStyle;
	int x = 0;
	int y = 0;
	UINT uWidth;
	UINT uHeight;

	switch( g_FACE.Mode )
	{
	case DEVIL_MODE_FULLSCREEN:
		dwExStyle = WS_EX_TOPMOST;
		dwStyle = WS_MAXIMIZE | WS_POPUP;
		x = 0;
		y = 0;
		uWidth   = GetSystemMetrics( SM_CXFULLSCREEN );
		uHeight  = GetSystemMetrics( SM_CXFULLSCREEN );

		g_bFullscreenQuitPermission = ( VerifyPassword == NULL );

		break;

	case DEVIL_MODE_PREVIEW:
		dwExStyle = 0;
		dwStyle = WS_CHILD;
		x = 0;
		y = 0;

		RECT ParentRect;
		GetWindowRect( g_hParent, &ParentRect );
		uWidth   = ParentRect.right - ParentRect.left;
		uHeight  = ParentRect.bottom - ParentRect.top;

		break;

	case DEVIL_MODE_WINDOWED:
		dwExStyle = WS_EX_WINDOWEDGE;
		dwStyle = WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_POPUP;

		int iScreenWidth   = GetSystemMetrics( SM_CXFULLSCREEN );
		int iScreenHeight  = GetSystemMetrics( SM_CYFULLSCREEN );
		uWidth	 = ( iScreenWidth  * 3 ) / 4;
		uHeight	 = ( iScreenHeight * 3 ) / 4;
		x = ( iScreenWidth  - uWidth  ) / 2;
		y = ( iScreenHeight - uHeight ) / 2;

		break;

	}


	// Switch resolution
	if( NULL != g_FACE.SwitchRes )
	{
		if( !g_FACE.SwitchRes( &x, &y, &uWidth, &uHeight ) )
		{
			return 0;
		}
	}


	// Create window
	g_FACE.hwnd = CreateWindowEx(
		dwExStyle,
		DEVIL_WINDOWCLASS,
		DEVIL_APPTITLE,
		dwStyle,
		x,
		y,
		uWidth,
		uHeight,
		g_hParent,
		NULL,
		g_hInstance,
		NULL
	);

	if( g_FACE.hwnd == NULL )
	{
		UnregisterClass( DEVIL_WINDOWCLASS, g_hInstance );
		return 0;
	}


	// OpenGL/D3D init
	if( NULL != g_FACE.Init3D )
	{
		if( !g_FACE.Init3D() )
		{
			UnregisterClass( DEVIL_WINDOWCLASS, g_hInstance );
			return 0;
		}
	}


	// Show window
	ShowWindow( g_FACE.hwnd, SW_SHOW );
	SetForegroundWindow( g_FACE.hwnd );
	if( g_FACE.Mode != DEVIL_MODE_PREVIEW )
	{
		SetFocus( g_FACE.hwnd );
	}


	// Message loop
	DEVIL_TIMER Timer;
	double fTimePassed = 0.0;
	MSG msg;
	msg.message = WM_QUIT + 1;			// Just to make it != WM_QUIT
	int iProcessed;
	while( msg.message != WM_QUIT )
	{
		iProcessed = 0;
		while( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
		{
			TranslateMessage( &msg );
			DispatchMessage( &msg );
			
			if( ++iProcessed >= 3 ) break;
		}

		if( g_bActive )
		{
			if( g_FACE.bEnableFrameLimiter )
			{
				if( 0 == g_FACE.uMaxFPS )
				{
					g_FACE.uMaxFPS = 40;
				}

				if( Timer.has_passed( 1000.0 / g_FACE.uMaxFPS, true, &fTimePassed ) )
				{
					g_FACE.fFPS = 1000.0 / fTimePassed;
					g_FACE.Render( fTimePassed );
				}
			}
			else
			{
				fTimePassed = Timer.get_passed_time( true );
				g_FACE.fFPS = 1000.0 / fTimePassed;
				g_FACE.Render( fTimePassed );
			}
		}

		Sleep( 1 );
	}


	// OpenGL/D3D cleanup
	if( g_FACE.Cleanup3D != NULL )
	{
		g_FACE.Cleanup3D();
	}

	
	// Unregister class
	UnregisterClass( DEVIL_WINDOWCLASS, g_hInstance );


	// Restore original resolution
	if( g_FACE.RestoreRes != NULL )
	{
		g_FACE.RestoreRes();
	}


	// Re-enable keys
	if( g_FACE.Mode == DEVIL_MODE_FULLSCREEN )
	{
#ifndef _DEBUG
		SystemParametersInfo( SPI_SCREENSAVERRUNNING, FALSE, NULL, 0 );
#endif

		DEVIL_UnloadPWD();
	}


	return 0;
}



// *****************************************************************************
int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, PSTR, int )
{
	g_hInstance = hInstance;

	// Deny execution of a unicode compile on a non-unicode system
	// Any component in here must be ansi!!!
	#ifdef UNICODE

		OSVERSIONINFOA osvi;
		osvi.dwOSVersionInfoSize = sizeof( OSVERSIONINFOA );
		GetVersionExA( &osvi );

		if( osvi.dwPlatformId != VER_PLATFORM_WIN32_NT )
		{
			MessageBoxExA( NULL, "This version was compiled for a UNICODE compatible system.\nThis operation system does not support UNICODE.\n\nPlease install the ANSI version.", "UNICODE error", MB_ICONERROR, MAKELANGID( LANG_ENGLISH, SUBLANG_ENGLISH_US ) );
			return 0;
		}

	#endif


	// Ensure we get enough CPU power
	SetPriorityClass( GetCurrentProcess(), NORMAL_PRIORITY_CLASS );

		
	// Get mode and parent window
	if( !DEVIL_GetMode() )
	{
		// We only get here if there's trouble with preview window handle
		return 0;
	}


	if( DEVIL_MODE_PASSWORD == g_FACE.Mode )
	{
		DEVIL_ChangePassword();
	}
	else
	{
		// Check interface
		if( !DEVIL_CheckInterface() )
		{
			return 0;
		}


		// Init
		if( NULL != g_FACE.Live )
		{
			if( !g_FACE.Live() )
			{
				return 0;
			}
		}


		// Run
		switch( g_FACE.Mode )
		{
		case DEVIL_MODE_CONFIG:
			if( g_FACE.Config != NULL )
			{
				g_FACE.Config( g_hParent );
			}
			else
			{
				DEVIL_DefaultConfig();
			}
			break;

		default:
			DEVIL_RunSaver();

		}
	

		// Cleanup
		if( NULL != g_FACE.Die )
		{
			g_FACE.Die();
		}
	}

	return 0;
}
